/*********************************************************************
	Component	: oxf
	Model Element	: RiJTimeoutManager
	Licensed Materials - Property of IBM
	(c) Copyright IBM Corporation 2008. All Rights Reserved.	
*********************************************************************/

package com.ibm.rational.rhapsody.oxf.timeouts;

import java.util.*;
//## operation getThread() 
import java.lang.Thread;
//## attribute unpostedTimeouts 
import java.util.TreeMap;
//## dependency oxf 
import com.ibm.rational.rhapsody.oxf.*;
//## dependency services 
import com.ibm.rational.rhapsody.oxf.services.*;
//## dependency AnimServices 
import com.ibm.rational.rhapsody.oxfinst.AnimServices;
//## link timer 
import com.ibm.rational.rhapsody.oxf.services.RiJTimer;
//## operation schedTimeout(RiJTimeout,RiJEventDispatcher) 
import com.ibm.rational.rhapsody.oxf.RiJEventDispatcher;
//## operation handleTimerInterrupt(RiJSimpleTimer) 
import com.ibm.rational.rhapsody.oxf.services.RiJSimpleTimer;
//## class RiJTimeoutManager 
import com.ibm.rational.rhapsody.oxf.services.RiJTimerListener;

//----------------------------------------------------------------------------
// com/ibm/rational/rhapsody/oxf/timeouts/RiJTimeoutManager.java                                                                  
//----------------------------------------------------------------------------

//## package com::ibm::rational::rhapsody::oxf::timeouts 


/**
This class counts time steps and posts timeouts to event dispatchers. It's a singleton.
*/
//## class RiJTimeoutManager 
public class RiJTimeoutManager implements RiJTimerListener {
    
    final public static long DEF_MAX_TIME = Long.MAX_VALUE-10000;		//## attribute DEF_MAX_TIME 
    
    final public static long DEF_TIME_TICK = 50;		//## attribute DEF_TIME_TICK 
    
    /**
    The time elapsed (milli seconds) since last reset.
    */
    protected long currentTime;		//## attribute currentTime 
    
    protected static RiJTimeoutManager m_instance;		//## attribute m_instance 
    
    /**
    The maximum value currentTime may have.
    */
    protected long maxTime = DEF_MAX_TIME;		//## attribute maxTime 
    
    /**
    Contains all timeouts that were inserted into their dispatcher but were not dispatched yet. It is the timeout responsibility to remove himself from this list after dispatching.
    Every timeout must have it's own unique id.
    */
    protected List postedTimeouts;		//## attribute postedTimeouts 
    
    protected boolean simulatedTime = false;		//## attribute simulatedTime 
    
    /**
    The time interval between notifyTime calls.
    */
    protected long timeTick = DEF_TIME_TICK;		//## attribute timeTick 
    
    /**
    This is a sorted map that contains all timeouts that are still waiting to be queued.
    */
    protected TreeMap unpostedTimeouts;		//## attribute unpostedTimeouts 
    
    /**
    the manager timer
    */
    protected RiJTimer timer;
    
    // Constructors
    
    // Argument boolean simulatedTime : 
    /**
    Indicates if time model is simulated time or not.
    */
    //## operation RiJTimeoutManager(long,long,boolean) 
    private  RiJTimeoutManager(long timeTick, long maxTime, boolean simulatedTime) {
        //#[ operation RiJTimeoutManager(long,long,boolean) 
        synchronized(this) {
        	currentTime = 0;         
        	if (timeTick != 0)
        		this.timeTick = timeTick;
        	if (maxTime != 0)
        		this.maxTime = maxTime;  
        	this.simulatedTime = simulatedTime;
        	RiJTimeoutComparator tC = new RiJTimeoutComparator();
        	unpostedTimeouts = new TreeMap(tC);
        	postedTimeouts = Collections.synchronizedList(new LinkedList());     
        
        	RiJTimerFactory timerFactory = RiJOXF.getTheTimerFactory();
        	if (simulatedTime) {
        		timer = timerFactory.createSimulatedTimeTimer(this);
        	}
        	else {
        		timer = timerFactory.createRealTimeTimer(this, this.timeTick, false);
        	}
        	if (timer != null)
        		timer.start();
        }
        
        //#]
    }
    /**
    this(DEF_TIME_TICK, DEF_MAX_TIME);
    */
    //## operation RiJTimeoutManager() 
    public  RiJTimeoutManager() {
        //#[ operation RiJTimeoutManager() 
        
        
        //#]
    }
    
    /**
    Consume the time in simulated time model. 
    */
    // Argument int interval : 
    /**
    Time interval to consume.
    */
    // Argument int step : 
    /**
    Steps to consume.
    */
    //## operation consumeTime(int,int) 
    public void consumeTime(int interval, int step) {
        //#[ operation consumeTime(int,int) 
        if (simulatedTime) {
            int phases = (int)(interval/step);
            int i;
            for (i=0;i<phases;i++) {
               synchronized(this) 
                {       
            	    currentTime += step;
            	    post(); // Always need to check for mature timeouts
                }
            }
        }    
        
        //#]
    }
    
    /**
    decrease reference counter of the simulated timer.
    */
    //## operation decNonIdleCount() 
    public void decNonIdleCount() {
        //#[ operation decNonIdleCount() 
        synchronized(this) {
          if (timer instanceof RiJSimulatedTimer)  
            ((RiJSimulatedTimer)timer).decNonIdleCount();   
        }
        
        //#]
    }
    
    /**
    find equal timeout in posted
    */
    //## operation findInPosted(RiJTimeout) 
    protected synchronized RiJTimeout findInPosted(RiJTimeout timeout) {
        //#[ operation findInPosted(RiJTimeout) 
        ListIterator li = postedTimeouts.listIterator();
        while (li.hasNext()) {
          RiJTimeout tm = (RiJTimeout)li.next();
          if (tm.equals(timeout)) return tm;
        }
        return null;
        //#]
    }
    
    /**
    Returns the thread that counts the ticks.
    */
    //## operation getThread() 
    public Thread getThread() {
        //#[ operation getThread() 
        Thread thread = null;
        if ((timer != null) && (timer instanceof RiJSimpleTimer)) {
          thread = ((RiJSimpleTimer)timer).getThread();
        }
        return thread;
        //#]
    }
    
    /**
    set the current time to the next scheduled timout expiration.
    */
    //## operation goNext() 
    protected void goNext() {
        //#[ operation goNext() 
        /*
        if ((unpostedTimeouts.isEmpty() == false) 
            && (postedTimeouts.isEmpty() == true))
        */
        if (unpostedTimeouts.isEmpty() == false) 
        {
            RiJTimeout timeout = (RiJTimeout)unpostedTimeouts.firstKey(); 
            currentTime = timeout.getDueTime();  
        }
        
        //#]
    }
    
    /**
    Sets the current time to the next pending timeout and posts it. This was provided so that "Go Idle" animation command will be able to advance the time in case the Event Queue is empty.
    */
    //## operation goNextAndPost() 
    public void goNextAndPost() {
        //#[ operation goNextAndPost() 
        goNext();
        post();
        //#]
    }
    
    /**
    Called by the Timer in case it was interrupted.
    */
    //## operation handleTimerInterrupt(RiJSimpleTimer) 
    public void handleTimerInterrupt(RiJSimpleTimer timer) {
        //#[ operation handleTimerInterrupt(RiJSimpleTimer) 
        //#]
    }
    
    /**
    increase reference counter of the simulated timer.
    */
    //## operation incNonIdleCount() 
    public void incNonIdleCount() {
        //#[ operation incNonIdleCount() 
        synchronized(this) {
          if (timer instanceof RiJSimulatedTimer)  
            ((RiJSimulatedTimer)timer).incNonIdleCount();
        }
        //#]
    }
    
    /**
    Increment current time. If maxTime reached reset absolute time of all timeouts.
    */
    //## operation incrementTime() 
    protected void incrementTime() {
        //#[ operation incrementTime() 
        // in simulated time the absolute time
        // is updated to the next timeout in the
        // list, instead of incrementing the time
        // tick by tick
        if (simulatedTime) 
        {
            goNext();       
        }
        else 
        {
            currentTime += timeTick;
        }
        
        if (currentTime >= maxTime) { 
          currentTime = 0;
          resetTimeouts();
        }
        //#]
    }
    
    /**
    This is for creating a new instance of RiJTimeoutManager. If an instance already exists the input is ignored.
    */
    //## operation instance(long,long) 
    public static RiJTimeoutManager instance(long tickTime, long maxTime) {
        //#[ operation instance(long,long) 
        return instance(tickTime, maxTime, false);
        //#]
    }
    
    /**
    get the singleton instance
    */
    //## operation instance() 
    public static RiJTimeoutManager instance() {
        //#[ operation instance() 
        return instance(RiJTimeoutManager.DEF_TIME_TICK, 
                        RiJTimeoutManager.DEF_MAX_TIME);
        //#]
    }
    
    /**
    This is for creating a new instance of RiJTimeoutManager. If an instance already exists the input is ignored.
    */
    // Argument boolean simulatedTime : 
    /**
    Indicates if time model is simulated time or not.
    */
    //## operation instance(long,long,boolean) 
    public static RiJTimeoutManager instance(long tickTime, long maxTime, boolean simulatedTime) {
        //#[ operation instance(long,long,boolean) 
        if (m_instance == null) {
          m_instance = new RiJTimeoutManager(tickTime, maxTime, simulatedTime);
        }
        return m_instance;
        //#]
    }
    
    //## operation isSimulatedTime() 
    public boolean isSimulatedTime() {
        //#[ operation isSimulatedTime() 
        return simulatedTime;
        //#]
    }
    
    /**
    This method is activated by the timer. The absolute time is promoted and expired time outs are posted.
    */
    //## operation notifyTime() 
    public void notifyTime() {
        //#[ operation notifyTime() 
        synchronized(this) {
          incrementTime();
          post();
        }
        
        //#]
    }
    
    /**
    post all expired timeouts
    */
    //## operation post() 
    protected void post() {
        //#[ operation post() 
        while (unpostedTimeouts.isEmpty() == false)
        {
            RiJTimeout timeout = (RiJTimeout)unpostedTimeouts.firstKey(); 
        
            // in case no timers expired break
            if (timeout.getDueTime() > currentTime) break;
        
            RiJEventDispatcher dispatcher = 
               (RiJEventDispatcher)unpostedTimeouts.get(timeout);
        
            // may need to add animation information
            RiJEvent ev = 
        	    AnimServices.initAnimEvent(timeout, timeout.getCreatingThread());
        
            dispatcher.queueEvent(ev);
        
            unpostedTimeouts.remove(timeout);
            postedTimeouts.add(timeout);
        }
        
        //#]
    }
    
    /**
    This method is to be used only from the animation framework. Its purpose is to allow incrementing the simulated time IF NECESSARY and queue the events for the next Go-Idle.  All of this is meant to avoid having to click multiple Go-Idle until simulated time is incremented.
    */
    //## operation promoteSimulatedTime() 
    public void promoteSimulatedTime() {
        //#[ operation promoteSimulatedTime() 
        if (simulatedTime && (timer instanceof RiJSimulatedTimer))
        {
            // check to see if time needs to be incremented
            RiJSimulatedTimer st = (RiJSimulatedTimer)timer;
            int rc = st.getNonIdleThreadsCounter();
            if (rc == 0) 
            {
                incrementTime(); 
                unsynchPost();   
            }
        }
        //#]
    }
    
    /**
    remove an expired timeout from postedTimeouts
    */
    //## operation removeExpiredTimeout(RiJTimeout) 
    protected synchronized void removeExpiredTimeout(RiJTimeout timeout) {
        //#[ operation removeExpiredTimeout(RiJTimeout) 
        postedTimeouts.remove(timeout);
        //#]
    }
    
    //## operation resetTimeouts() 
    protected void resetTimeouts() {
        //#[ operation resetTimeouts() 
        // only the unposted ones need resetting
        Iterator iter = unpostedTimeouts.keySet().iterator();
        while (iter.hasNext()) {   
          ((RiJTimeout)(iter.next())).resetDueTime(currentTime);
        }
        //#]
    }
    
    /**
    Schedules a timeout event to be queued into the dispacther.
    */
    //## operation schedTimeout(RiJTimeout,RiJEventDispatcher) 
    public void schedTimeout(RiJTimeout timeout, RiJEventDispatcher dispatcher) {
        //#[ operation schedTimeout(RiJTimeout,RiJEventDispatcher) 
        synchronized(this)
        {
            timeout.resetDueTime(currentTime);
            // register as the timerManager of the timeout.
            timeout.setTimeoutManager(this);
            
            // inserts timeout and dispatcher into the unpostedTimouts map
            unpostedTimeouts.put(timeout, dispatcher);
        }
        //#]
    }
    
    /**
    set simulated time mode. 
    setSimulatedTime(true) - time model is set to simulated time.
    setSimulatedTime(false) - time model is set to real time.
    */
    //## operation setSimulatedTime(boolean) 
    public void setSimulatedTime(boolean simulateTime) {
        //#[ operation setSimulatedTime(boolean) 
        // reset timeouts and absolute time
        currentTime = 0;
        resetTimeouts();
        
        simulatedTime = simulateTime;
        //#]
    }
    
    /**
    Stopes the timer thread.
    */
    //## operation stopTimer() 
    public void stopTimer() {
        //#[ operation stopTimer() 
         if (timer != null)
         	timer.stop(); 
          
        //#]
    }
    
    /**
    unsched a timeout (posted and unposted)
    */
    //## operation unschedTimeout(RiJTimeout) 
    public void unschedTimeout(RiJTimeout theTimeout) {
        //#[ operation unschedTimeout(RiJTimeout) 
        synchronized (this)
        {
            // if posted get theRiJEventDispatcher and cancel the event
            int index;
            RiJTimeout tm;
            if ((index = postedTimeouts.indexOf(theTimeout)) >= 0) 
            {
              tm = (RiJTimeout)postedTimeouts.get(index);
              postedTimeouts.remove(index);
              return;
            }
            else { // if not posted, remove and foerget it all !
              if (unpostedTimeouts.remove(theTimeout) == null) { // relying on equals method
              	//may happen that the TreeMap 'unpostedTimeouts' fails to remove a key since the dummy tm (theTimeout)
        		//wasn't assigned for its 'dueTime', and the compare function in the RiJTimeoutComparator uses the 'dueTime'
        		//to decide how to sort the timeouts.
        		//a simple workaround is to traverse the TreeMap, find the Timeout, and remove it
        		Iterator iter = unpostedTimeouts.keySet().iterator();
        		RiJTimeout cancelTm = null;
        		while (iter.hasNext() ) {   
        			tm = (RiJTimeout)iter.next();
        			if(tm.equals(theTimeout))
        			{
        				cancelTm = tm;
        				break;
        			}	
        		}
        		if(cancelTm != null) {
        			unpostedTimeouts.remove(cancelTm);
        		}
              }
            }
        }
        //#]
    }
    
    /**
    post all expired timeouts in an unsychronized way. 
    To be used by animation framework alone.
    */
    //## operation unsynchPost() 
    protected void unsynchPost() {
        //#[ operation unsynchPost() 
        while (unpostedTimeouts.isEmpty() == false)
        {
            RiJTimeout timeout = (RiJTimeout)unpostedTimeouts.firstKey(); 
        
            // in case no timers expired break
            if (timeout.getDueTime() > currentTime) break;
        
            try 
            {    
                RiJThread dispatcher = 
                   (RiJThread)unpostedTimeouts.get(timeout);
                
                // may need to add animation information
                RiJEvent ev = 
                    AnimServices.initAnimEvent(timeout, timeout.getCreatingThread());
                
                dispatcher.unsynchQueueEvent(ev);
                
                unpostedTimeouts.remove(timeout);
                postedTimeouts.add(timeout);
            }
            catch (ClassCastException ce) {
            }
        }
        
        //#]
    }
    
    public long getCurrentTime() {
        return currentTime;
    }
    
}

