/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2008. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/

package docgen;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.nio.channels.FileChannel;

import com.ibm.rational.rpe.api.docgen.IDocumentGenerator;
import com.ibm.rational.rpe.api.docgen.LocalDocumentGenerator;
import com.ibm.rational.rpe.api.docspec.DocumentSpecificationBuilder;
import com.ibm.rational.rpe.api.docspec.RPEDocumentSpecification;
import com.ibm.rational.rpe.api.docspec.RPEListener;
import com.ibm.rational.rpe.api.docspec.RPEResult;
import com.ibm.rational.rpe.common.config.RPEConfigConstants;
import com.ibm.rational.rpe.common.defs.BuildInfo;
import com.ibm.rational.rpe.common.utils.AppUtils;
import com.ibm.rational.rpe.common.utils.Credential;
import com.ibm.rational.rpe.common.utils.PropertyUtils;
import com.ibm.rational.rpe.common.utils.RPEException;
import com.ibm.rational.rpe.common.utils.URIUtils;
import com.ibm.rational.rpe.engine.RRDGEngine;
import com.ibm.rational.rpe.engine.RRDGEngine.EngineStatus;
import com.ibm.rhapsody.apps.utils.Reporter;

public class SampleDocgen 
{
	//private static SampleLog4JLogger logger = null;
	private IDocumentGenerator generator = null;
	private RPEDocumentSpecification docspec = null;
	private DocumentSpecificationBuilder builder = null;

	private static class ConsoleEcho implements RPEListener 
	{
		// @Override
		public void progress(String message) 
		{
			Reporter.report(message);
		}
	}
	
	public boolean isInitialized() {
		return docspec != null;
	}
	
	public void initialize(String docSpecURL, String productHome, String workArea, Credential credential, boolean isDebugMode) throws RPEException,URISyntaxException, IOException, InterruptedException 
	{
		RRDGEngine engine = new RRDGEngineImpl();
		// enable debug mode only if needed
		engine.setDebugMode(isDebugMode);				

		generator = new LocalDocumentGenerator(engine);
		
		// set this first as RPE needs to pick various files from this location. You can bundle this with your app but you need to make it available 
		// on the file system for RPE to access it.The example contains a "RPE" folder with everything that is needed. 
		// Copy that folder on your file system and change the path bellow to match it
		if ( productHome.length() != 0)
		{
			AppUtils.setProductHome( productHome);
		}
		
		// set the location where RPE will create output files and temporary files. If not specified RPE will use the user's temp folder
		if ( workArea.length() != 0)
		{
			AppUtils.setProductWorkArea( workArea);
		}	
				
		builder = new DocumentSpecificationBuilder();

		// the configuration file must be provided to load operations. AppUtils provides the method to find its location based on product home. 
		// Alternatively you can specify its path explicitly
		URI configURI = new File( AppUtils.getConfigFilePath()).toURI();

		// register a listener for RPE's engine messages. Note that these messages are a subset of what RPE logs in its log file
		generator.registerListener(new ConsoleEcho( ));
		
		Reporter.report("Loading document specification ..."); //$NON-NLS-1$
		
		// load the doc spec
		docspec = builder.load(configURI.toString(), URIUtils.encodeURI(docSpecURL));
	}

	public int publish(String generatedDocument) throws RPEException, IOException {

		// perform the document generation synchronously
		EngineStatus resultStatus = generator.publishSync( docspec, 0);

		// if document generation was successful display the results
		if ( resultStatus == EngineStatus.finished) 
		{
			Reporter.report("Engine execution finished succesfully. The results are:"); //$NON-NLS-1$
			
			for( RPEResult resultFile : generator.getResults())
			{
				String location = PropertyUtils.getPropertyRawValue( resultFile.getProperty(RPEConfigConstants.PROPERTY_PATH), "");  //$NON-NLS-1$
				String type = PropertyUtils.getPropertyRawValue( resultFile.getProperty(RPEConfigConstants.PROPERTY_TYPE), "");  //$NON-NLS-1$
				
				Reporter.report("\t" + type + ": " + location); //$NON-NLS-1$ //$NON-NLS-2$
				
				copyFile(location, generatedDocument);
			}
		}
		
		return resultStatus == EngineStatus.finished ? 0 : 1;
	}
	
	public static void copyFile(String src, String dest) throws IOException {
		if (!src.equals("") && !dest.equals("")) {
			
			File sourceFile = new File(src);
			File destFile = new File(dest);
			
			if(!destFile.exists()) {
		        destFile.createNewFile();
		    }
	
		    FileChannel source = null;
		    FileChannel destination = null;
	
		    try {
		        source = new FileInputStream(sourceFile).getChannel();
		        destination = new FileOutputStream(destFile).getChannel();
		        destination.transferFrom(source, 0, source.size());
		    }
		    finally {
		        if(source != null) {
		            source.close();
		        }
		        if(destination != null) {
		            destination.close();
		        }
		    }
		}
	}

	public static void main(String[] args) 
	{
		configureLogging("");
				
		if (args.length < 1) 
		{
			System.exit( 1);
			return;
		}

		File docSpecFile = new File(args[0]);
		if (!docSpecFile.exists()) 
		{
			Reporter.report("Docspec cannot be found at location: " + args[0]); //$NON-NLS-1$
			System.exit( 1);
			return;
		}
		
		String specFile = args[0];
		String productHome = ""; //$NON-NLS-1$
		String workArea = ""; //$NON-NLS-1$
		String generatedDocument = "";
		boolean isDebugMode = false;
				
		int pos = 1;
		while( pos < args.length - 1)
		{
			if ( args[pos].equalsIgnoreCase( "-home")) //$NON-NLS-1$
			{
				productHome = args[pos+1];
				pos += 2;
			} 
			else if ( args[pos].equalsIgnoreCase( "-workArea")) //$NON-NLS-1$
			{
				workArea = args[pos+1];
				pos += 2;
			}
			else if ( args[pos].equalsIgnoreCase( "-generatedDocument")) //$NON-NLS-1$
			{
				generatedDocument = args[pos+1];
				pos += 2;
			}
			else if ( args[pos].equalsIgnoreCase( "-debugMode")) //$NON-NLS-1$
			{
				isDebugMode = Boolean.valueOf(args[pos+1]);
				pos += 2;
			}				
			else
			{
				//logger.error(" Unknown argument " + args[pos]); //$NON-NLS-1$
				System.exit( 1);
				return;
			}
		}
		
		SampleDocgen sampleDocgen = new SampleDocgen();
		int exitCode = sampleDocgen.initialize(specFile, productHome, workArea, isDebugMode);
		if (exitCode != 1) {
			sampleDocgen.generateReport(generatedDocument);
		}
		
		System.exit( exitCode);
	}

	public int initialize(String specFile, String productHome, String workArea, boolean isDebugMode) {
		if (isInitialized()) {
			return 0;
			
		}
		String docSpec = new File(specFile).toURI().toString();
		
		Reporter.report("");
		Reporter.report("Running document specification: " + docSpec); //$NON-NLS-1$
		Reporter.report("RPE Engine version: " + BuildInfo.getEngineVersion()); //$NON-NLS-1$
		Reporter.report("Java version: " + System.getProperty("java.version")); //$NON-NLS-1$ //$NON-NLS-2$
		Reporter.report("Debug Mode: " + String.valueOf(isDebugMode)); //$NON-NLS-1$ //$NON-NLS-2$
		
		// use the built configuration if none provided. This only works in the dev environment and when using the built jar
		// you need to provide this location in the "home" argument 
		if ( (productHome == null) || productHome.isEmpty())
		{
			String base = SampleDocgen.class.getProtectionDomain().getCodeSource().getLocation().getPath();
			productHome = new File( base, "home").getAbsolutePath(); //$NON-NLS-1$
		}

		int exitCode = 1;
		
		try 
		{
			initialize( docSpec, productHome, workArea, new Credential("", ""), isDebugMode);
		} 
		catch (RPEException e) 
		{
			AppUtils.getLogger().error( e.getLocalizedMessage(), e);
		}
		catch (URISyntaxException e) 
		{
			AppUtils.getLogger().error( e.getLocalizedMessage(), e);
		}
		catch ( IOException e)
		{
			AppUtils.getLogger().error( e.getLocalizedMessage(), e);
		}
		catch (InterruptedException e)
		{
			AppUtils.getLogger().error( e.getLocalizedMessage(), e);
		}
		
		Reporter.report("Initialization done."); //$NON-NLS-1$
		return exitCode;
	}
	
	public int generateReport(String generatedDocument) {
		int exitCode = 1;
		
		try 
		{
			exitCode = publish(generatedDocument);
		} 
		catch (RPEException e) 
		{
			AppUtils.getLogger().error( e.getLocalizedMessage(), e);
		}
		catch ( IOException e)
		{
			AppUtils.getLogger().error( e.getLocalizedMessage(), e);
		}
		
		Reporter.report("Report generation done."); //$NON-NLS-1$
		return exitCode;
	}
	
	
	public static void configureLogging(String path) 
	{
		//PropertyConfigurator.configure( SampleDocgen.class.getClassLoader().getResource("log4j.properties")); //$NON-NLS-1$
		//PropertyConfigurator.configure( new File(path).toURI().toString()); //$NON-NLS-1$
		//logger = new SampleLog4JLogger();
	}
}
